/*******************************************************************************
 * MagSWv20a Magnetic switch for R/C models
 * Module: MAIN
 * Author: Jiri Panacek, jiri@panacek.net, www.panacek.net
 * Version: 1.0.1 19-FEB-2019
 * License: Open-source, public domain
 * 
 * Change log:
 * v1.0.1 19-FEB-2019: First official release (no changes to v1.0.1-RC)
 * v1.0.1-RC 09-FEB-2019: Optimized sleep consumption (16uA -> 3.5uA)
 * v1.0-RC 28-NOV-2019: First working version, internal testing only
 *******************************************************************************/
#define _XTAL_FREQ 8000000

#include <xc.h>
#include <stdint.h>
#include <pic10f322.h>

/* The CONFIG word - copy and paste from the in-built MPLABX generator */
#pragma config FOSC = INTOSC    // Oscillator Selection bits (INTOSC oscillator: CLKIN function disabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#pragma config PWRTE = ON       // Power-up Timer Enable bit (PWRT enabled)
#pragma config MCLRE = OFF      // MCLR Pin Function Select bit (MCLR pin function is digital input, MCLR internally tied to VDD)
#pragma config CP = OFF         // Code Protection bit (Program memory code protection is disabled)
#pragma config LVP = OFF        // Low-Voltage Programming Enable (Low-voltage programming disabled)
#pragma config LPBOR = OFF      // Brown-out Reset Selection bits (BOR disabled)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)

/* notes: 
 * # disable LVP in order to allow for internal MCLR 
 * # disable BOR to reduce current consumption during SLEEP
 */


/*******************************************************************************
 * MACROS
 ******************************************************************************/

/* GPIOs mapping */
#define LED     PORTAbits.RA1   /* RA1 drives the LED (pos) */
#define EN      PORTAbits.RA2   /* RA2 drives the high-side switch (pos) */
#define TRIG_N  PORTAbits.RA3   /* RA3/MCLR is the trigger (hall) input (neg) */

/* Thresholds for the voltage measurement */
#define ADC_LEVEL_HIGH  N/A     /* High Level >= 4.84V */ 
#define ADC_LEVEL_MID   120     /* Medium Level < 4.84V */
#define ADC_LEVEL_LOW   100     /* Low Level < 4.03V */
#define ADC_LEVEL_CRIT  87      /* Critical Level < 3.51V */

/* Hall sensor trigger signal */
#define HALL_DEGLITCH_PERIOD_MS 100 /* Minimum pulse-duration for on/off */ 

/* Led signalization behavior, each bit represents 100ms interval */
#define LED_PATTERN_HIGH    0xFF /* /--------/ Constant ON */
#define LED_PATTERN_MID     0xEF /* /.-------/ Single blink and ON */
#define LED_PATTERN_LOW     0xAF /* /.-.-----/ Double blink and ON */
#define LED_PATTERN_CRIT    0xAA /* /.-.-.-.-/ Continously blinking */

/*******************************************************************************
 * ENUMS
 ******************************************************************************/
enum {
    FSM_RESET, FSM_ON, FSM_OFF
}; /* System Finite State Machine (FSM) states */

enum {
    STANDBY, TRIGGERED
}; /* The event for the FSM */

/*******************************************************************************
 * TYPEDEFS
 ******************************************************************************/

/*******************************************************************************
 * VARIABLES, CONSTANTS
 ******************************************************************************/

volatile uint8_t AdcMinVal = 255; /* The lowest detected voltage */
volatile uint8_t HallDeglitchTmr = 0; /* De-glitch SW timer for the hall sens.*/
volatile uint8_t TriggerFlag = STANDBY; /* Flag informing a valid trigger */

/*******************************************************************************
 * INTERNAL FUNCTIONS PROTOTYPES
 ******************************************************************************/

inline void SystemCfg(void); /* MCU H/W configuration */
inline void LedService(void); /* The func. serves blinking of the LED */

/*******************************************************************************
 * INTERNAL FUNCTIONS
 ******************************************************************************/

/* MCU H/W configuration */
inline void SystemCfg(void) {
    ANSELA = 0x01; /* RA0 is an analog input */
    TRISA = 0x09; /* RA0, RA3 inputs & RA1, RA2 outputs */
    ADCON = 0xA0; /* ADC off, AN0, Fosc/16 */
    INTCON = 0xC8; /* Enable global, peripheral and on-change interrupts */
    IOCAN = 0x08; /* Int. on-change enabled for a negative going edge on RA3 */
    IOCAP = 0x08; /* Int. on-change enabled for a positive going edge on RA3 */
    PIE1 = 0x02; /* Enable interrupt from TMR2 */
    PR2 = 124; /* 8Mhz clock & 1/16 prescale & 125 period causes 1ms ISR */
    T2CON = 0x06; /* Prescaler 1/16, enable timer */
    VREGCONbits.VREGPM1 = 1; /* SUPER IMPORTANT - sets sleep consumption 
                              * to 3.5 uA instead of 16uA ! */
        
}

/* The func. serves blinking of the LED */
inline void LedService(void) {
    static uint8_t LedIdx = 0; /* bit position index */
    static uint8_t LedMode;
  
    /* Identify the voltage level & assign the LED behavior */
    if (AdcMinVal < ADC_LEVEL_CRIT) {
        LedMode = LED_PATTERN_CRIT;
    } else if (AdcMinVal < ADC_LEVEL_LOW) {
        LedMode = LED_PATTERN_LOW;
    } else if (AdcMinVal < ADC_LEVEL_MID) {
        LedMode = LED_PATTERN_MID;
    } else {
        LedMode = LED_PATTERN_HIGH;
    }
    /* ON/OFF the LED with regards to the bit */
    LED = (0x01 & (LedMode >> LedIdx));
    __delay_ms(100);
    /* Increment the bit position index or start from zero */
    if (LedIdx < 7) LedIdx++;
    else LedIdx = 0;
}

/*******************************************************************************
 * MAIN FUNCTION
 ******************************************************************************/
void main(void) {
    SystemCfg();

    while (1) {
        /* finite state machine */
        static uint8_t fsm_state = FSM_RESET;
        static uint8_t fsm_last_state = !(FSM_RESET);
        static uint8_t fsm_next_state = !(FSM_RESET);

        switch (fsm_state) {
            case FSM_RESET:
            {
                /* Entry + Exit Action */
                EN = 1; /* Start-always-on */
                LED = 1;
                fsm_next_state = FSM_ON;
                break;
            }
            case FSM_ON:
            {
                /* Entry Action */
                if (fsm_state != fsm_last_state) {
                    EN = 1;
                    /* ADC Enable + delay */
                    __delay_ms(1); /* The eFuse output has some slew-rate.. */
                    ADRES = 0xFF; /* Reset the ADC vale and run the ADC */
                    AdcMinVal = 0xFF;
                    ADCONbits.ADON = 1;
                }
                /* Led Signalization */
                LedService();                         
                /* Exit Action */
                if (TRIGGERED == TriggerFlag) {
                    TriggerFlag = STANDBY;
                    fsm_next_state = FSM_OFF;
                }
                break;
            }
            case FSM_OFF:
            {
                /* Entry Action */
                if (fsm_state != fsm_last_state) {
                    EN = 0; /* Turn off the output */
                    LED = 0;
                    ADCONbits.ADON = 0; /* Stop the ADC */
                }
                SLEEP();
                /* Exit Action */
                while (HallDeglitchTmr) {
                };
                if (TRIGGERED == TriggerFlag) {
                    TriggerFlag = STANDBY;
                    fsm_next_state = FSM_ON;
                }
                break;
            }
            default:
            {
                /* Should theoretically never happen */
                RESET(); /* But if happens (e.g. x-ray). reset the system */
                break;
            }
        }
        fsm_last_state = fsm_state; /* Register the FSM state change */
        fsm_state = fsm_next_state; /* if any */
    }
    return;
}

/*******************************************************************************
 * INTERRUPTS
 ******************************************************************************/
void __interrupt() isr(void) {
    /*
     *  TIMER2 interrupt (every 1ms) 
     */
    if ((PIR1bits.TMR2IF)&&(PIE1bits.TMR2IE)) {
        if (ADCONbits.ADON == 1) {
            /* Process the previous ADC sample */
            if (ADRES < AdcMinVal) {
                AdcMinVal = ADRES;
            }
            /* Trigger the ADC (every 1ms if ON) */
            ADCONbits.GO_nDONE = 1;
        }
        /* Service de-glitch timer */
        if (HallDeglitchTmr > 0) {
            --HallDeglitchTmr;
            if (HallDeglitchTmr == 0) {
                TriggerFlag = TRIGGERED; /* Set the flag for the FSM */
            }
        }
        PIR1bits.TMR2IF = 0; /* clear TMR2 interrupt flag */
    }

    /* 
     * RA3 on-change interrupt 
     */
    if (IOCAFbits.IOCAF3) {
        /* Falling Edge */
        if (TRIG_N == 0) {
            HallDeglitchTmr = HALL_DEGLITCH_PERIOD_MS; /* Start the SW timer */
        }/* Rising edge*/
        else {
            HallDeglitchTmr = 0; /* Any rising edge disables the timer */
        }
        IOCAFbits.IOCAF3 = 0; /* clear on-change interrupt flag for RA3 */
    }
}
/*******************************************************************************
 * NOTES:
 *  o   PIC10F322 is an extremely small microcontroller with limited memory. For 
 *      this reason, the source code does not use encapsulation. Additionally,
 *      the code uses -inline- functions and global variables when necessary. 
 * 
 *  o   Modifying the source code for ATTiny10 is probably possible. It has 
 *      similar peripheral features and is pin-compatible.
 * 
 *  o   The 1.0RC version takes 60% of FLASH and 28% of RAM with the free XC8 
 *      compiler (v2.00).
 * 
 *  o   The output of the LDO (TPS709) is 1% accurate. This is significantly
 *      higher accuracy than the internal FVR reference. God knows why Microchip
 *      bothered with implementing a voltage reference with +6/-8% accuracy.
 *      Additionally, the FVR can't be used as a reference to the ADC, the ADC
 *      can just measure it. This theoretically allows for a single or two-point
 *      calibration but why?? For this reason, I don't do any calibration.
 * 
 * OPERATION:
 *  1.  The code configures GPIOs, interrupt on-change, ADC and TIMER2 after 
 *      the reset.
 * 
 *  2.  The MCU executes the finite state machine (FSM) in the endless 'while' 
 *      loop. The FSM has three states (RESET, ON, OFF). The RESET state is 
 *      rather unnecessary but present for completeness. 
 * 
 *  3.  The FSM enters the RESET and then the ON state automatically. This 
 *      implements the always-turn-on behavior.
 * 
 *  4.  The ON state enables the ADC. The TIMER2 overflows every 1ms and 
 *      triggers the ADC - if enabled. Additionally, the routine services a SW
 *      timer which de-glitches the trigger input.
 *    
 *  5.  A falling edge on the trigger input causes an interrupt which starts the 
 *      SW timer.  Any rising edge stops the SW timer. The timer routine sets 
 *      the TriggerFlag to TRIGGERED if the pulse is longer than the de-glitch 
 *      period.
 * 
 *  6.  The FSM checks for the TriggerFlag and changes the state from ON to OFF
 *      or back if the system detected a valid pulse.
 * 
 *  7.  Once the FSM enters the OFF state, the MCU switches the output off and 
 *      enters the low-power (SLEEP) mode. Any change on the trigger input 
 *      wakes the MCU. If the pulse is not valid, the MCU goes to sleep again.
 * 
 *  8.  The system measures the voltage on the output, remembers the lowest 
 *      value and reports it by blinking the LED. The minimal value resets 
 *      every on/off.
 *  
 *  9.  The LED service routine walks repeatedly through the each bit of a 
 *      variable with the size of a byte. The LED goes on/off with regards to 
 *      the bit. The duration of every bit is approx. 100ms. This allows for 
 *      different blinking patterns. 
 *      Example: 0b10101010 is blinking 1:1 with a change every 100ms, 
 *      b1100110011 is also blinking 1:1 but with a change every 200ms etc...
 *  
 *      note: I chose the FSM concept after experimenting with simple if-else. 
 *      It turned out that it is more flexible (wake-up, delays, LEDs...).
 * 
 * Used H/W resources:
 *  o   8-b ADC
 *  o   Timer2 which causes interrupt every 1ms
 *  o   Interrupt on-change (rising+falling edge) for TRIG_N signal (RA3)
 *  o   Sleep
 * 
 ******************************************************************************/
